import sys
import getpass
import secrets
import os
import os.path
import random
from cryptography.fernet import Fernet
from base64 import urlsafe_b64encode as b64e, urlsafe_b64decode as b64d

from cryptography.hazmat.backends  import default_backend as backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

import rsa
from rsa import RsaKey

DEFAULT_DATABASE_PATH = "db.txt"
USAGES = {
        "new": 
"new <mode> [ <key> ] <length> [ <label> <character set> ]\
\n        'mode' = csr or rsa, mode of encryption\
\n        'key' = the key to use when making a caesar cipher, must be a number. Not necessary for RSA!\
\n        'length' = num characters in password \
\n        'label' = optional name for the password \
\n        'charset' = characters that can be in the password (use '-<characters>' to exclude a list of characters))\
",
        "strength": "strength <password/label> - Evaluates the strength of a password, or if a label is provided then that password.",
        "echo": "echo <string content> ( 'string_content' = any text to be echoed )",
        "dbg": "dbg ( Prints a debug output of the manager's state )",
        "list": "list [ <label> <password> ](Prints the passwords currently stored 'label' = the password you want to decrypt)",
        "save": "save <filename> - Saves the state to the file originally inputted or the file provided.",
        "drop": "drop <label> - deletes the selected label from the list of passwords. This can be undone with the 'undo' command, however if 'drop' is used on an already deleted password, the password is removed forever from the database.",
        "undo": "undo - undoes the latest deletion/modification",
        "history": "history - prints all the deleted passwords",
        "cfg": "config [ <param=newparam> ] - configures the current program state or lists it when run without arguments. Example: 'config hints=False'",
        "rename": "rename <label> <new label> - Changes a password's label",
}
DEFAULT_KEYGEN_ITERATIONS = 100_000
LOWERCASE_ALPHABET = set([chr(e) for e in range(ord('a'), ord('a') + 25)])
UPPERCASE_ALPHABET = set([chr(e) for e in range(ord('A'), ord('A') + 25)])
NUMBERS = set([chr(e) for e in range(ord('0'), ord('0') + 10)])
SPECIAL = set([chr(e) for e in range(ord('!'), ord('!') + 14)]) ^ set(',') # no commas in our passwords, sorry
CHARSET_MAP = {
        "l": LOWERCASE_ALPHABET,
        "u": UPPERCASE_ALPHABET,
        "n": NUMBERS,
        "s": SPECIAL,
}

SAFEST = 50

"""
Simple function that prints a prompt message to get a confirmation message.
"""
def confirm(msg, yes, no):
    i = input(msg + f" ({yes}/{no})")
    i.strip()
    s = i.lower()
    if s == yes:
        return True
    elif s == no:
        return False
    else:
        return False

def derive_key(password: bytes, salt: bytes, iterations: int = DEFAULT_KEYGEN_ITERATIONS) -> bytes:
    kdf = PBKDF2HMAC(
            algorithm = hashes.SHA256(), length=32, salt=salt,
            iterations=iterations, backend=backend)
    return b64e(kdf.derive(password))

def password_encrypt(message: bytes, password: str, iterations: int = DEFAULT_KEYGEN_ITERATIONS) -> bytes:
    salt = secrets.token_bytes(16)
    key = derive_key(password.encode(), salt, iterations)
    return b64e(
            b'%b%b%b' % (
                salt,
                iterations.to_bytes(4, 'big'),
                b64d(Fernet(key).encrypt(message)),
            )
    )

def password_decrypt(token: bytes, password: str) -> (bytes, bytes):
    decoded = b64d(token)
    salt, iters, token = decoded[:16], decoded[16:20], b64e(decoded[20:])
    iterations = int.from_bytes(iters, 'big')
    key = derive_key(password.encode(), salt, iterations)
    return (key, Fernet(key).decrypt(token))

def encrypt(message: bytes, key: bytes) -> bytes:
    return Fernet(key).encrypt(message)

def decrypt(token: bytes, key: bytes) -> bytes:
    return Fernet(key).decrypt(token)

"""
Password class contains:
    - key => The key used to encrypt and decrypt the password, 'int' for caesar, 'Tuple(bigint, bigint)' for RSA
    - mode => 'rsa' or 'csr', the encryption method used for the password
    - password => the unencrypted password

Note: it is probably a bad idea to have the program open on a multi-user device, as the passwords
are stored in Python interpreter memory unobfuscated, and as such can be found using simple C function calls. 
Memory obfuscation is possibly a feature to add in the future or to use as the consumer of this program.
"""
class Password:
    def __init__(self, password, key, mode):
        self.password = password
        self.key = key
        self.mode = mode

    def __str__(self):
        return f"{self.key}, {self.mode}, {repr(self.password)}"

    def __repr__(self):
        return self.__str__()

    """
    Accept a key and an encrypted password and return the decrypted password. 

    NOTE: this uses an unbeleivably dangerous loophole with 'eval' that should 
    not be allowed to release into the final version
    """
    def decrypt(password, key, mode):
        if mode == "csr":
            s = []
            for l in password:
                s.append(chr(ord(l) - int(key)))
            return "".join(s)
        if mode == "rsa":
            # inSANE SECURITY VULNERABILITY 
            # fix: figure out how to turn b64 hex byte string into bytes object
            # DO NOT LET THIS GO INTO PROD!!!!!!!!!!!!!!!!!!!!!!!!
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            password = eval(password)
            return rsa.decrypt_data(key.q, password)

    """
    Returns the encrypted bytes of the password depending on password type.
    """
    def encrypt(self):
        match self.mode:
            case "rsa":
                return self.rsa_encrypt()
            case "csr":
                return self.csr_encrypt()
            case _:
                raise Exception

    """Returns serialized string representing this password for saving.
    """
    def serialize(self):
        return f"{self.key},{self.encrypt()}"

    """
    Encrypt a password with helper functions from the rsa.py module.
    """
    def rsa_encrypt(self):
        return rsa.encrypt_data(self.key.p, self.password)

    """
    Increment the index of each character by one.
    """
    def csr_encrypt(self):
        s = []
        for l in str(self.password):
            s.append(chr(ord(l) + int(self.key)))
        return "".join(s)

"""
State of password manager, passwords, keys, and path of database file.

NOTE: the class's constructor is also its load function.
"""
class State:
    def __init__(self, filepath, password):
        if not os.path.isfile(filepath):
            raise FileNotFoundError
        self.filepath = filepath
        try:
            file = open(filepath, 'r')
        except:
            raise FileNotFoundError

        content = file.read()
        lines = content.split('\n')
        
        head = lines[0]
        try:
            self.sha_key, decrypt = password_decrypt(head, password)
        except:
            print("Password does not match for this database, and as such it cannot be decrypted.")
            raise Exception

        tail = lines[1]
        s = tail.split(":")
        self.rsa_key = RsaKey(eval(s[0]),eval(s[1]))

        self.passwords = {}
        self.to_remove = []
        self.overwritten = []
        self.head = head

        if content == "":
            file.close()
            return

        mode = None
        key = None

        for i, line in enumerate(lines):
            if i < 2:
                continue
            if line == "":
                continue

            if 'mode:' in line:
                parts = line.split(':')

                if len(parts) != 3:
                    print(f"Error parsing mode switch on line {i}: {line}\n ^^^^ -> Expected three parts, got {len(parts)}")
                    continue

                mode = parts[1]
                key = parts[2]
                if mode == 'rsa':
                    key = self.rsa_key
                continue

            # Expect: key,label,encrypted_password
            parts = line.split(",", 1)
            if len(parts) != 2:
                print(f"Error parsing password on line {i}: {line}\n ^^^^ -> Expected two parts, got {len(parts)}.")
                continue

            if mode is None or key is None:
                print(f"Error parsing password on line {i}: {line}\n ^^^^ -> Missing mode description above.")
                continue

            # Expect: old(label) or rm(label)
            l = parts[0].split("(")
            if len(l) != 1 and l[0] == "rm":
                label = parts[0].split("(")[1][:-1] # trim ending parenthesis
                
                self.passwords[label] = Password(Password.decrypt(parts[1], key, mode), key, mode)
                
                self.to_remove.append(label)
            elif len(l) != 1 and l[0] == "old":
                self.overwritten.append(l[1][:-1])
                self.passwords[parts[0]] = Password(Password.decrypt(parts[1], key, mode), key, mode)
            else:
                self.passwords[parts[0]] = Password(Password.decrypt(parts[1], key, mode), key, mode)

        file.close()

    """
    Add a password to the state depending on what type of password it is.

    NOTE: an empty string as a label will cause the label to be the number of passwords + 1
    """
    def add_password(self, label, password_in, mode, key=None):
        if label is None:
            label = str(len(self.passwords) + 1)

        if self.rsa_key is None:
            print("TODO: generate RSA key")

        if key is None:
            p = Password(password_in, self.rsa_key, mode)
            self.passwords[label] = p
            return

        p = Password(password_in, key, mode)
        self.passwords[label] = p


    """
    Returns password as bytes.
    """
    def get_password(self, label):
        if label in self.passwords:
            p = self.passwords[label]
            if p.mode == "csr":
                return str(p.password)[2:-1]
            else:
                return p.password.decode()
        else:
            return b"[ERROR: no such label found]"

    """
    Saves the State to State.filepath.

    NOTE: the database file currently contains no version data, making incompatibility un unnoticed error.

    FORMAT:
    <head> <-- verification that inputted password is correct using Fernet encryption
    <rsa key> <-- RSA key corresponding to this state
    <foreach p in passwords>
        if p is caesar:
            'mode:csr:<p.key>'                  <-- each caesar password has its own key
            '<p.label>,<p.encrypted_password>'  <-- comma separated
    <end>
    'mode:rsa:_' <-- _ means the key is defined at the top of the file
    <foreach p in passwords>
        if p is rsa:
            '<p.label>,<p.encrypted_password>'  <-- same format as caesar, just in a different section
    <end>
    """
    def save(self, filepath = None):
        if filepath is None:
            filepath = self.filepath
        content = self.head + "\n"
        content += f"{self.rsa_key}" + "\n"
        for l, p in self.passwords.items():
            if p.mode == "csr":
                content += f"mode:csr:{p.key}\n"
                if l in self.to_remove:
                    content += f"rm({l}),{p.encrypt()}\n"
                else:
                    content += f"{l},{p.encrypt()}\n"
        if self.rsa_key is not None:
            content += f"mode:rsa:_\n"
        for l, p in self.passwords.items():
            if p.mode == "rsa":
                if l in self.to_remove:
                    content += f"rm({l}),{p.encrypt()}\n"
                else:
                    content += f"{l},{p.encrypt()}\n"
        file = open(filepath, 'w')
        file.write(content)
        file.close()

"""
Generate a password given a 'set' object of characters to choose from.
"""
def generate_password(length, charset):
    s = []
    for i in range(0, length):
        s.append(random.choice(list(charset)))
    return ''.join(s)

def handle_args(db: str, password: str):
    state = State(db, password)
    return state

def main():
    filepath = ""
    password = ""
    if len(sys.argv) < 2:
        print(f"No database path supplied. Usage: {sys.argv[0]} <database path> <password>")
        return
    else:
        filepath = sys.argv[1]

    if len(sys.argv) < 3:
        if os.path.isfile(filepath):
            print(f"No password supplied. Usage: {sys.argv[0]} <database path> <password>")
            return
        else:
            password = "???"
    else:
        password = sys.argv[2]

    try:
        state = handle_args(filepath, password)
    except FileNotFoundError:
        print(f"Error while opening database file, {filepath} does not exist or cannot be opened.")
        ask = input(f"Create {filepath}? (Y/N)")
        ask = ask.strip()
        ask = ask.lower()
        if ask == "y":
            password = ""
            while True:
                # TODO: make this print asterisks or waddeva
                ask = input(f"Please input a password for {filepath} (Note: password cannot contain spaces): ")
                ask = ask.strip()
                if ask == "q":
                    print("Cancelling.")
                    return
                elif len(ask) > 100 or len(ask) < 8 or ask == "":
                    print("Password is too long/short, please try again.")
                else:
                    print("Type password again to confirm:")
                    conf = input(f"Confirmation: ")
                    conf = conf.strip()
                    if conf != ask:
                        print("Error, passwords do not match.")
                        continue
                    break

            password = ask
            file = open(filepath, "xb")
            key = rsa.generate_rsa_key();
            keysha = password_encrypt("super secret hihi >_>".encode(), password)
            file.write(keysha)
            file.write('\n'.encode())
            file.write(f"{key}".encode())
            file.write('\n'.encode())
            file.close()
            return 
        else:
            return
    except Exception as e:
        print(f"Error parsing database {filepath}: {e}")
        return

    cfg = {
        "hints": True
    }
    print("Welcome to SmashPass. Type 'h' or 'help' for help")

    while True:
        command_input = input(">")
        command_input = command_input.strip()

        args = command_input.split(" ")
        if len(args) == 0:
            continue

        command = args[0]

        if command == "q" or command == "quit" or command == "exit":
            if len(args) == 2:
                if args[1] == "f" or args[1] == "force" or args[1] == "nosave":
                    if confirm("Are you sure you want to exit without saving?", "y", "n"):
                        return
            else:
                break
        if command == "cfg" or command == "config" or command == "settings":
            if len(args) < 2:
                print(f"Settings: hints: {cfg['hints']}, ...")
                continue

            s = args[1].split("=")
            if len(s) < 2:
                print(f"Incorrect format for modifying program parameters. Usage: {USAGES['cfg']}")
                continue

            if s[0] not in cfg:
                print(f"Unknown cfg parameter '{s[0]}'. Type 'cfg' for a list of config parameters.")
                continue
            try:
                cfg[s[0]] = s[1]
            except:
                print(f"Incorrect format for cfg parameter {s[0]}, expected type {type(cfg[s[0]])}")
        elif command == "help" or command == "h" or command == "?":
            if len(args) == 1:
                print("Available commands: ")
                print("\t" + ", ".join(USAGES.keys()))
                if cfg['hints'] == True: 
                    print("(Hint: type 'help <command name>' for more detailed info about that command)")
            else:
                cmd = args[1]
                if not cmd in USAGES:
                    print("Unknown command, you may have mistyped it.")
                    continue
                print(USAGES[cmd])
        elif command == "echo":
            if len(args) < 2:
                print(f"Too few arguments for {command}, usage: {USAGES['echo']}") 
                continue
            print(" ".join(args[1:]))
        elif command == "dbg":
            print(vars(state))
        elif command == "save" or command == "s":
            file = None
            if len(args) > 1:
                file = args[1]
            try:
                state.save(file)
                print(f"State saved to {file if file is not None else state.filepath}!")
            except:
                print("There was an error saving, maybe back up your stuff now. Sorry")
        elif command == "drop" or command == "delete" or command == "del" or command == "rm":
            if len(args) < 2:
                print(f"Too few arguments for command {command}, usage: {USAGES['drop']}")
                continue
            
            label = args[1].strip().lower()
            if label in state.to_remove:
                if confirm(f"Are you sure you want to PERMANANTLY DELETE {label}? FOREVER????? (Note: this CANNOT be undone)", "y", "n"):
                    del state.passwords[label]
                    del state.to_remove[state.to_remove.index(label)]
                    print(f"Password '{label}' was permanently removed from database. This cannot be undone.")
            elif label in state.passwords:
                if confirm(f"Are you sure you want to delete {label}?", "y", "n"):
                    state.to_remove.append(label)
                    print(f"Password '{label}' removed from database.")
                    if cfg['hints'] == True:
                        print("(Hint: use 'undo' to undo this operation, and 'history' to see previous deleted passwords and restore them)")
            else:
                print(f"Password '{label}' was not found, have you spelt it correctly?")
        elif command == "undo" or command == "un":
            if len(state.to_remove) == 0:
                print("Already on latest change")
                continue
            ltst = state.to_remove.pop()
            print(f"Undone latest deletion: {ltst}")
        elif command == "rename" or command == "replace" or command == "ren":
            if len(args) < 3:
                print(f"Too few arguments for command {command}. Usage: {USAGES['rename']}")
                continue
            
            label = args[1].strip().lower()
            if label not in state.passwords:
                print(f"Password {args[1]} not found in database.")
            else:
                newlabel = args[2].strip().lower()
                if newlabel in state.passwords:
                    print(f"Cannot rename to {newlabel}, as there is already a password with that name")
                    continue
                state.passwords[newlabel] = state.passwords[label]
                del state.passwords[label]
                print(f"Renamed {label} to {newlabel}")
        elif command == "restore" or command == "res" or command == "r":
            if len(args) < 2:
                print(f"Too few arguments for command {command}. Usage: {USAGES['restore']}")
                continue

            args[1].strip()
            og_label = args[1].lower()
            label = og_label
            if og_label in state.passwords and (og_label in state.overwritten or og_label in state.to_remove):
                print("You will have to rename this password to restore it.")
                label_in = input("New label (If left empty, a number will be used): ")
                if label_in == "":
                    label = str(len(state.passwords) + 1)
                else:
                    label_in.strip()
                    label = label_in.lower()

            if og_label in state.to_remove:
                del state.to_remove[state.to_remove.index(og_label)]
                print(f"Restored {og_label}")
            elif og_label in state.overwritten:
                del state.overwritten[state.overwritten.index(og_label)]
                state.passwords[label] = state.passwords[f"old({og_label})"]
                del state.passwords[f"old({og_label})"]
                print(f"Restored {og_label}")
            else:
                print(f"Label '{og_label}' could not be found in deleted or overwritten passwords")
        elif command == "history" or command == "his" or command == "undolist":
            print("Deleted:")
            for remd in state.to_remove:
                print(f"{remd}: [...]", end=' ')
            print("\nOverwritten:")
            for over in state.overwritten:
                print(f"{over}: [...]", end=' ')
            print("")

        elif command == "strength" or command == "strchk" or command == 'howstrong':
            if len(args) < 2:
                print(f"Too few arguments for {command}. Usage: {USAGES['strength']}")
                continue

            pword = args[1]

            if args[1] in state.passwords:
                pword = str(state.passwords[args[1]].password)

            # 'p' means percent (or coefficient)
            # 'd' = diversity 
            lp = len(pword) 
            dp = len([ele for ele in SPECIAL if (ele in pword)])
            dp += len([ele for ele in NUMBERS if (ele in pword)])
            total_safety = lp + dp
            percent = total_safety / SAFEST
            if percent > 1.0:
                print("Passwords exceeds max safety expected, perfect!")
            else:
                print(f"[{'*' * int(percent * 15)}{'_' * int((1-percent) * 15)}]")
        elif command == "new": # Form: 0:new 1:<mode> 2:<key> 3:<length> [ 4:<label> ]
            if len(args) < 2:
                print(f"Too few arguments for {command}, usage: {USAGES['new']}") 
                continue

            label = None
            s_length = None
            key = None
            charset = "all"

            mode = args[1]
            if mode == 'csr':
                if len(args) < 4:
                    print(f"Too few arguments for new caesar password. Usage: {USAGES['new']}")
                    continue
                key = args[2]
                s_length = args[3]

                if len(args) > 4:
                    label = args[4]
                if len(args) > 5:
                    charset = args[5]
            elif mode == 'rsa':
                if len(args) < 3:
                    print(f"Too few arguments for new rsa password: Usage: {USAGES['new']}");
                    continue
                key = None
                s_length = args[2]
                if len(args) > 3:
                    label = args[3]
                if len(args) > 4:
                    charset = args[4]
            else:
                print(f"Mode '{mode}' is unknown. Modes are 'rsa' and 'csr'")
                continue

            try:
                length = int(s_length)
            except:
                print(f"Error parsing length as number, usage: {USAGES['new']}")
                continue


            if label in state.passwords:
                print(f"Error generating new password, label '{label}' is already in use.")
                if label in state.overwritten:
                    if confim("Do you want to PERMANENTLY OVERWRITE {label}? (THIS CANNOT BE UNDONE)", "y", "n"):
                        todo
                if confirm("Would you like to replace it?", "y", "n"):
                    state.overwritten.append(label)
                    state.passwords[f"old({label})"] = state.passwords[label]
                    print(f"Overwriting '{label}' with new password...")
                    if cfg['hints'] == True:
                        print("(Hint: use the 'restore' command to undo this action)")
                else:
                    continue

            chars = set()

            if charset == "all":
                chars = LOWERCASE_ALPHABET | UPPERCASE_ALPHABET | NUMBERS | SPECIAL
            else:
                s = charset.split("-")
                for c in list(s[0]):
                    chars |= CHARSET_MAP[c]
                if len(s) > 1:
                    for c in s[1]: # list of chars to exclude
                        chars ^= set(c)

            password = generate_password(length, chars)
            print(password)
            state.add_password(label, password.encode(), mode, key)
        elif command == "list" or command == "ls":
            if len(args) == 1:
                # TODO: make it possible to scroll thru a list of passwords
                for l in state.passwords:
                    if l not in state.to_remove and "old" not in l:
                        print(f"{l}: [...]")
                if cfg['hints'] == True:
                    print("(Hint: type 'list <key>' to decrypt your chosen password")
            elif len(args) == 2:
                label = args[1].strip().lower()
                print(state.get_password(args[1]))
            else:
                print(f"Too few args for {command}. Usage: {USAGES['list']}")

            

    state.save()
    print("Thanks for using!")


if __name__ == "__main__":
    main()
